﻿//Star Micronics 2016 - C# SDK Line Mode StarIO.
//Please note the word document included with this SDK
//It contains more detailed information on this SDK and StarIO
using System;
using System.Globalization;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading;
//////////////Attach the StarIOPort.dll in the project and include its methods.
using StarMicronics.StarIO;
////////////////////////////////////////////////////////////////////////

namespace Sample_Receipt_Csharp_StarIO
{
    partial class SMForm
    {
        //General purpose delegate
        delegate void Action();
        //Delegate for searching the received data from the printer
        delegate int DataSearch(List<byte> datas, byte[] searchValue);

        //Time-out value when transmitting and receiving data with the printer
        const int TIMEOUT = 1000;
        //Buffer value at the time of receiving data from the printer
        const int RECEIVE_BUFFER_SIZE = 256;

        //Flag to receive the bar code data from repeat printer
        bool readingFlag = false;
        //Thread for receiving the bar code data
        Thread _thread = null;

        //Dictionary for replacing control characters of the ASCII code to a string
        private Dictionary<Byte, String> asciiCodeDic = new Dictionary<Byte, String>
        {
            { 0x00, "<NUL>" },
            { 0x01, "<SOH>" },
            { 0x02, "<STX>" },
            { 0x03, "<ETX>" },
            { 0x04, "<EOT>" },
            { 0x05, "<ENQ>" },
            { 0x06, "<ACK>" },
            { 0x07, "<BEL>" },
            { 0x08, "<BS>"  },
            { 0x09, "<TAB>" },
            { 0x0A, "<LF>"  },
            { 0x0B, "<VT>"  },
            { 0x0C, "<FF>"  },
            { 0x0D, "<CR>"  },
            { 0x0E, "<SO>"  },
            { 0x0F, "<SI>"  },
            { 0x10, "<DLE>" },
            { 0x11, "<DC1>" },
            { 0x12, "<DC2>" },
            { 0x13, "<DC3>" },
            { 0x14, "<DC4>" },
            { 0x15, "<NAK>" },
            { 0x16, "<SYN>" },
            { 0x17, "<ETB>" },
            { 0x18, "<CAN>" },
            { 0x19, "<EM>"  },
            { 0x1A, "<SUB>" },
            { 0x1B, "<ESC>" },
            { 0x1C, "<FS>"  },
            { 0x1D, "<GS>"  },
            { 0x1E, "<RS>"  },
            { 0x1F, "<US>"  },
        };

        /// <summary>
        /// Methods when the reading Start / Stop button of the bar code data is pressed
        /// </summary>
        private void cmdReadBarcodes_Click(object sender, EventArgs e)
        {
            cmdReadBarcodes.Text = "Please wait ...";
            cmdReadBarcodes.Enabled = false;

            if (_thread == null || !_thread.IsAlive)
            {
                //If the bar code receiving thread is not live, to start the thread
                readingFlag = true;

                _thread = new Thread(BarcodeReader);
                _thread.Name = "Barcode Reader Thread";
                _thread.Start();
            }
            else
            {
                //If the bar code receiving thread is alive, it will terminate the reading loop of bar code data
                readingFlag = false;
            }
        }

        /// <summary>
        /// Reading processing of bar code (another thread)
        /// </summary>
        private void BarcodeReader()
        {
            //Update button on the label
            Action<String> _UpdateButtonLabel = (text) =>
                {
                    cmdReadBarcodes.Text = text;
                };
            Action<String> _UpdateLblPrinterStatus = (text) =>
                {
                    lblPrinterStatus.Text = text;
                    lblPrinterStatus.ForeColor = Color.Green;
                };
            Action _CmdReadBarcodesStop = () =>
                {
                    cmdReadBarcodes.Text = "Stop";
                    cmdReadBarcodes.Enabled = true;
                };
            //In the byte array, to search whether the specified byte column exists
            DataSearch _DataSearch = (datas, searchValue) =>
                {
                    int index = datas.IndexOf(searchValue[0]);

                    while (index > -1)
                    {
                        bool match = true;

                        for (int i = 1; i < searchValue.Length; i++)
                        {
                            if (datas.Count < index + i || datas[index + i] != searchValue[i])
                            {
                                match = false;
                                break;
                            }
                        }

                        if (match)
                        {
                            break;
                        }

                        index = datas.IndexOf(searchValue[0], index + 1);
                    }

                    return index;
                };
            Action<UInt32> _CountShow = (count) =>
                {
                    labelCountValue.Text = count.ToString();
                };
            //Via the main thread anonymous method for adding the log
            Action<Byte[]> _AppendLog = (data) =>
                {
                    appendLog(data);
                };
            //To display the data received from the bar code reader
            Action<Byte[]> _AppendBcrLog = (data) =>
                {
                    this.fromBcrData.Text = "";

                    StringBuilder sb = new StringBuilder();
                    sb.Capacity = data.Length * 5;

                    if (this.fromBcrDataHex.Checked)
                    {
                        //Hex
                        
                        foreach (byte b in data)
                        {
                            sb.Append("0x");
                            sb.Append(b.ToString("X2"));
                            sb.Append(' ');
                        }
                    }
                    else
                    {
                        //ASCII

                        //Encode the string to ASCII
                        Encoding encoder = ASCIIEncoding.GetEncoding("us-ascii", new EncoderReplacementFallback(String.Empty), new DecoderExceptionFallback());
                        Char[] encodedBytes = encoder.GetChars(data);

                        foreach (char b in encodedBytes)
                        {
                            //Check whether the control characters in the dictionary
                            if (asciiCodeDic.ContainsKey((byte)b))
                            {
                                //If it was a control character, to bring instead of a string
                                sb.Append(asciiCodeDic[(byte)b]);
                            }
                            else
                            {
                                sb.Append(b);
                            }

                            sb.Append(' ');
                        }
                    }

                    this.fromBcrData.Text = sb.ToString();
                };
            Action _CmdReadBarcodesStart = () =>
                {
                    cmdReadBarcodes.Text = "Start";
                    cmdReadBarcodes.Enabled = true;
                };
            //Initialization of when an exception occurs
            Action<String> _ExceptionExit = (string text) =>
                {
                    ClosePort();

                    lblPrinterStatus.Text = text;
                    lblPrinterStatus.ForeColor = Color.Red;

                    _CmdReadBarcodesStart();
                };


            //List in which to keep the command to be transmitted to the printer
            byte[] commands;
            //List in which to keep data received from the printer or a bar code reader
            List<byte> readDatas = new List<byte>();


            //Update button on the label
            this.Invoke(_UpdateButtonLabel, "Please wait ...");

            
            //Set the status to offline because this is a new attempt to print
            this.onlineStatus = false;

            //TRY -> Use the textboxes to check if the port info given will connect to the printer.
            try
            {
                //Very important to only try opening the port in a Try, Catch incase the port is not working
                if(txtPortName.Text.ToLower().Contains("bt:"))
                {
                    this.sPort = Factory.I.GetPort(txtPortName.Text, txtPortSetting.Text, 2000);
                }
                else
                {
                    this.sPort = Factory.I.GetPort(txtPortName.Text, txtPortSetting.Text, 10000);
                }

                //GetOnlineStatus() will return a boolean to let us know if the printer was reachable.
                this.onlineStatus = sPort.GetOnlineStatus();
            }

            //CATCH -> If the port information is bad, catch the failure.
            catch (PortException)
            {
                this.Invoke(_ExceptionExit, "Port Error");
                return; //exit this entire function
            }
            
            //If it is offline, dont setup receipt or try to write to the port.
            if (this.onlineStatus == false)
            {
                this.Invoke(_ExceptionExit, "Offline");
                return; //exit this entire function
            }

            //Else statement means it is ONLINE, lets start the printjob  
            this.Invoke(_UpdateLblPrinterStatus, "Online");



            this.Invoke(_UpdateButtonLabel, "Buffer cleaning now...");


            //Clear the buffer of the bar code at the printer
            //                      <ESC> <GS>   'B'   '3'
            commands = new byte[] { 0x1B, 0x1D, 0x42, 0x33 };
            this.Invoke(_AppendLog, commands);

            //Send commands to the printer
            if (!Send(commands))
            {
                //Sending of command failure
                this.Invoke(_ExceptionExit, "Buffer clear command send Error");
                return;
            }


            this.Invoke(_UpdateButtonLabel, "Getting barcode reader status ...");

            //Receive connection status of the bar code reader
            //                      <ESC> <GS>   'B'   '1'
            commands = new byte[] { 0x1B, 0x1D, 0x42, 0x31 };

            this.Invoke(_AppendLog, commands);

            if (!Send(commands))
            {
                this.Invoke(_ExceptionExit, "Get barcode reader status command send Error");
                return;
            }

            //Flag is confirmed that the bar code reader is connected
            bool bcrConnection = false;

            //For the detection of time-out, and save the operating time of the computer
            int tickCount = Environment.TickCount;

            //Stop button has not been pressed, and if no foam is closed,
            //                                          and the connection of the bar code reader is not able to detect, and repeats
            while (readingFlag && !this.formClosing && !bcrConnection)
            {
                //It receives the data, to save
                readDatas = Receive();

                //The analysis of the connection status of the bar code reader
                //
                //Reply contents of the command
                // <ESC> <GS> 'B' '1' n
                //                  n : 0x00 = No connection of barcode reader
                //                      0x01 = No connection of barcode reader and Data exists in barcode buffer
                //                      0x02 = Barcode reader is connected to the printer
                //                      0x03 = Barcode reader is connected to the printer and Data exists in barcode buffer
                
                //                                      <ESC> <GS>   'B'   '1'   connected
                if (_DataSearch(readDatas, new byte[] { 0x1B, 0x1D, 0x42, 0x31, 0x02 }) > -1)
                {
                    //Barcode reader is connected to the printer
                    bcrConnection = true;
                }

                if (!bcrConnection)
                {
                    if (Environment.TickCount - tickCount > TIMEOUT)
                    {
                        //Timeout
                        this.Invoke(_ExceptionExit, "Timeout");
                        return;
                    }
                }
            }


            if (bcrConnection)
            {
                //After changing the label of the button to "Stop", to enable the button
                this.Invoke(_CmdReadBarcodesStop);


                //Bar code reader has read, it will request a bar code data
                //                      <ESC> <GS>   'B'   '2'
                commands = new byte[] { 0x1B, 0x1D, 0x42, 0x32 };

                this.Invoke(_AppendLog, commands);

                //It counted the number of times the data has been received from the printer
                uint loopCount = 0;

                //Stop button has not been pressed, and if no foam is closed,
                //                                          and the other methods unless sending data to the printer, and repeats
                while (readingFlag && !this.formClosing && this.onlineStatus)
                {
                    try
                    {
                        this.Invoke(_CountShow, ++loopCount);

                        Send(commands);

                        readDatas = Receive();

                        //Analysis of the received barcode data
                        //
                        //Reply contents of the command
                        // <ESC> <GS> 'B' '2' n1 n2 d1 d2 ... dx
                        //                                    n1 n2 : Express the length of the bar code data
                        //                                              barcode data length = "n1 + (n2 * 256)" bytes
                        //                                    d     : Barcode datas

                        //Identify the starting position of the received bar code data
                        //                                                   <ESC> <GS>   'B'   '2'
                        int startIndex = _DataSearch(readDatas, new byte[] { 0x1B, 0x1D, 0x42, 0x32 });

                        if (startIndex > -1)
                        {
                            //Discard request command portion
                            startIndex += 4; //<ESC> <GS> 'B' '2'

                            //The analysis of the length of the bar code data
                            int barcodeDataLength = readDatas[startIndex] + (readDatas[startIndex + 1] * 256);
                            //Data at the end of 2byte is <CR> <LF>
                            barcodeDataLength = barcodeDataLength - 2;

                            if (barcodeDataLength < 1)
                            {
                                //Bar code data does not exist
                                continue;
                            }

                            //Discard length data of the bar code data
                            startIndex += 2; //n1 n2

                            byte[] barcodeDatas = new byte[barcodeDataLength];

                            for (int i = 0; i < barcodeDataLength && startIndex + i < readDatas.Count; i++)
                            {
                                byte data = readDatas[startIndex + i];
                                barcodeDatas[i] = data;
                            }

                            //Add the bar code data in the log
                            this.Invoke(_AppendBcrLog, barcodeDatas);
                        }
                    }
                    catch
                    {
                        //It will ignore even exception occurs
                    }
                }
            }


            ClosePort();

            if (!this.formClosing)
            {
                //If no form is closed, initialize the label of the button
                this.Invoke(_CmdReadBarcodesStart);
            }
        }

        /// <summary>
        /// Send commands to the printer
        /// </summary>
        /// <param name="buffer">Send commands array</param>
        /// <returns>Status was to be sent to the printer</returns>
        private bool Send(byte[] buffer)
        {
            //Write bytes to printer
            uint amountWritten = 0;
            uint amountWrittenKeep;
            while (buffer.Length > amountWritten)
            {
                //This while loop is very important because in here is where the bytes are being sent out through StarIO to the printer
                amountWrittenKeep = amountWritten;
                try
                {
                    amountWritten += sPort.WritePort(buffer, amountWritten, (uint)buffer.Length - amountWritten);
                }
                catch (PortException)
                {
                    return false; //exit this entire function
                }

                if (amountWrittenKeep == amountWritten) // no data is sent
                {
                    return false; //exit this entire function
                }
            }

            return true;
        }

        /// <summary>
        /// Receive data from the printer
        /// </summary>
        /// <returns>The received data</returns>
        private List<byte> Receive()
        {
            //Array to keep the received data
            List<byte> receiveBuffer = new List<byte>();

            //Array that temporarily stores the received data
            byte[] buffer = new byte[] { }; 

            //For the detection of time-out, and save the operating time of the computer
            int tickCount = Environment.TickCount;

            //Between the received data is empty, receives repeatedly
            while (receiveBuffer.Count == 0)
            {
                try
                {
                    //To resize the array to the length to be received
                    //Adding 8, remnants of the Windows API
                    Array.Resize(ref buffer, RECEIVE_BUFFER_SIZE + 8);

                    //It receives the data, to save the actual length of the data
                    uint realLength = this.sPort.ReadPort(ref buffer, 0, RECEIVE_BUFFER_SIZE);

                    //Resize the array to the length of the received data
                    Array.Resize(ref buffer, (int)realLength);

                    receiveBuffer.AddRange(buffer);
                }
                catch
                {
                }

                if (receiveBuffer.Count == 0)
                {
                    if (Environment.TickCount - tickCount > TIMEOUT)
                    {
                        //Timeout
                        break;
                    }
                }
            }

            return receiveBuffer;
        }

        /// <summary>
        /// If the port was open, close the port
        /// </summary>
        private void ClosePort()
        {
            if (this.sPort != null)
            {
                Factory.I.ReleasePort(this.sPort);
            } 
        }
    }
}